import glob
import netCDF4
import os
import struct
from .space  import SpaceList
import netcdf
#import pp

def _file_format(openfile):
    '''

Private function to read a file's magic number.

Takes a file object as input and assumes that we're currently
pointing to the beginning of the file. The file is rewound to the
beginning after reading the magic number.

Returns a string giving the format of the input file.
''' 

    if openfile.name.startswith('http://'):
        # Assume that a URL is netCDF
        return 'nc'

    # Read the magic number
    magic_number = struct.unpack('=L', openfile.read(4))[0]

    # Reset the pointer to the beginning of the file
    openfile.seek(0, os.SEEK_SET)

    # netCDF
    if magic_number in (21382211, 1128547841, 38159427): #1128547841
        return 'nc'

    # PP
    if magic_number in (256, 65536):
        raise NotImplementedError, "PP input is not implemented yet"
        return 'pp'

    # Add more file formats here ...

    # Still here?
    raise NotImplementedError, "Unsupported file format from %s. Magic number=%d" % \
        (openfile.name, magic_number)
#--- End: def

def _read_a_file(filename, ignore_ioerror=False, verbose=False,
                 close=False,
                 extract={}):

    # Find this input file's format
    if filename.startswith('http://'):
        format = 'nc'
    else:        
        try:
            openfile = open(filename, 'rb') 
        except IOError as error:
            if ignore_ioerror:
                print 'Ignoring IOError: %(error)s' % \
                    locals()
                return SpaceList()
            raise IOError, error
                
        format = _file_format(openfile)        
    
    # ----------------------------------------------------
    # Read the file into spaces
    # ----------------------------------------------------
    if format == 'nc':
        spaces = netcdf.read(filename, close=close, extract=extract)
        
    elif format == 'pp': 
        raise NotImplementedError, "PP input is not implemented yet"
        spaces = pp.read(openfile, close=close)

    # Add more file formats here ...

    return spaces
#--- End: def

def read(files, read_data=False, verbose=False, index=None, 
         ignore_ioerror=False, close=False, 
         calendar=None,
         long_name=None,
         standard_name=None,
         units=None,
         extract={},
         ):
    '''

Read spaces from input files from disk or from an OPeNDAP server.

By default, all data arrays from all space components are not read
from disk but are stored in the returned spaces as file pointers. This
behaviour may be changed with the `read_data` parameter.

NetCDF dimension names are stored in the `nc_dimensions` attribute of
a space's grid and netCDF variable names are stored in the private
attribute `ncvar` of the space and each of its grid components.

Spaces referenced by formula_terms transforms are not included in the
returned list of spaces. A space may be referenced in the transforms
of multiple output spaces, so in-place changes to that space in one
transform will be seen in the other spaces which reference it.

:argument files: A string or sequence of strings giving the file names
                 or OPenDAP URLs from which to read spaces. For files
                 on disk, the file names may contain wild cards as
                 understood by the :mod:`glob` module.

:type files: `str` or sequence

:parameter read_data: *Optional*. If True then read the data of each
                      space, including the data of its grid components
                      into memory Otherwise all data are stored as
                      file pointers. data.

:type data: `bool`

:parameter index: *Optional*. If None then return all possible spaces
                  from the input files. If a non-negative integer then
                  only return the space with this index in the
                  otherwise full list of spaces.

:type index: `None` or `int`

:parameter verbose: *Optional*. If True then print information to
                    stdout.

:type verbose: `bool`

:parameter close: *Optional*. If True then close each file after it
                  has been read.

:type verbose: `bool`

:parameter ignore_ioerror: *Optional*. If True then ignore files which
                           produce IO errors (such as empty files, for
                           example).

:type verbose: `bool`

:parameter calendar: *Optional*. Equivalent to setting 'calendar' in
                     the `extract` keyword dictionary

:type calendar: `str`


:parameter long_name: *Optional*. Equivalent to setting 'long_name' in
                      the `extract` keyword dictionary

:type long_name: `str`


:parameter standard_name: *Optional*. Equivalent to setting
                          'standard_name' in the `extract` keyword
                          dictionary

:type standard_name: `str`


:parameter units: *Optional*. Equivalent to setting 'units' in the
                  `extract` keyword dictionary

:type units: `str`

:parameter extract: *Optional*. A dictionary of space selection
                    criteria. Only spaces which match the criteria are
                    returned. Refer to :meth:`~cf.Space.ematch` for
                    details.

:type extract: `dict`

:returns: A list of spaces in a :class:`~cf.SpaceList` object.

**Examples:**

>>> s = cf.read('file*.nc')
>>> type(s)
<class 'cf.space.SpaceList'>
>>> s
[<CF Space: pmsl(30, 24)>,
 <CF Space: z-squared(17, 30, 24)>,
 <CF Space: temperature(17, 30, 24)>,
 <CF Space: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc')[0:2]
[<CF Space: pmsl(30, 24)>,
 <CF Space: z-squared(17, 30, 24)>]

>>> cf.read('file*.nc', index=0)
[<CF Space: pmsl(30, 24)>]

>>> cf.read('file*.nc')[-1]
<CF Space: temperature_wind(17, 29, 24)>

>>> cf.read('file*.nc', extract={'units': 'K'})
[<CF Space: temperature(17, 30, 24)>,
 <CF Space: temperature_wind(17, 29, 24)>]

>>> cf.read('file*.nc', standard_name='.*pmsl*', extract={'units': 'K|Pa'})[0]
<CF Space: pmsl(30, 24)>
'''

    # Initialize the output list of spaces
    space_list = SpaceList()

    # Initialize the extract dictionary
    extract = extract.copy()
    if calendar is not None:
        extract['calendar']      = calendar    
    if long_name is not None:
        extract['long_name']     = long_name
    if standard_name is not None:
        extract['standard_name'] = standard_name
    if units is not None:
        extract['units']         = units

    # Count the number of spaces (in all files) and the number of
    # files
    space_counter = -1
    file_counter  = 0

    if isinstance(files, basestring):
        files = (files,)

    for file_glob in files:

        if file_glob.startswith('http://'):
            # Do not glob a URL
            files2 = (file_glob,)
            url = True
        else:
            # Do glob files on disk
            files2 = glob.iglob(file_glob)
            url = False

        for filename in files2:

            # Print some informative messages
            if verbose and index is None:
                print 'File: %(filename)s' % locals()

            # ----------------------------------------------------
            # Read the file into spaces
            # ----------------------------------------------------
            spaces = _read_a_file(filename,
                                  ignore_ioerror=ignore_ioerror,
                                  verbose=verbose,
                                  close=close,
                                  extract=extract)
            
            # ------------------------------------------------------
            # Add this file's spaces to those alread read from other
            # files
            # ------------------------------------------------------
            space_list.extend(spaces)
   
            space_counter = len(space_list)
            file_counter += 1

            # Print some informative messages
            if verbose and index is None:
                i = space_counter - len(spaces)
                for s in spaces:
                    print '%d: %s' % (i, repr(s))
                    i += 1
            #--- Endf: if

            # ------------------------------------------------------
            # If we only want one space from all input files then
            # break now if we have got it
            # ------------------------------------------------------
            if index is not None and space_counter >= index+1:
                break
        #--- End: for
            
        # -----------------------------------------------------------
        # If we only one space from all input files then break now if
        # we have got it
        # -----------------------------------------------------------
        if index is not None and space_counter >= index:
            break
    #--- End: for     

    # Error check
    if not ignore_ioerror:
        if not file_counter:
            raise RuntimeError('No files found')
        if not space_list:
            raise RuntimeError('No spaces found from '+str(file_counter)+' files')
    #--- End: if
    if index >= len(space_list):
        raise IndexError('SpaceList index='+str(index)+' is out of range')

    # Print some informative messages
    if verbose:
        if index is None:  
            print 'Read %d space%s from %d file%s' % \
                (space_counter, ('s' if space_counter!=1 else ''),
                 file_counter , ('s' if file_counter !=1 else ''))
        else:
            print 'File: %(filename)s' % locals()
            print '%d: %s' % (0, repr(space_list[index]))
            print 'Read 1 space from 1 file'
    #--- End: if
    
    # ------------------------------------------------------------
    # Read the data, if requested, and return the final list of
    # spaces, or a single space if index has been set
    # ------------------------------------------------------------
    try:
        space_list = SpaceList(space_list[index])
    except TypeError:
        pass

    if read_data:
        space_list.slice[...]

    return space_list
#--- End: def
