import netCDF4
import copy
import re
import numpy
from .space import Space, SpaceList, Variable
from .utils import equals

# ----------------- ------------------------------------------------------
# Global variable   Description
# ----------------- ------------------------------------------------------
# netcdf            netCDF4.Dataset instance
#
# nc                Dictionary relating netCDF variable names and
#                   netCDF4.Variable instances
#
# dim_to_ncdim      Dictionary relating space dimension names and
#                   netCDF dimension names
#
# ncdim_size        Dictionary relating netCDF dimension names and
#                   their sizes
#
# seen              Dictionary relating elements of the space (such as
#                   a coordinate or a transform) to their netCDF
#                   variable names and their netCDF dimensions
#
# ncvar_names       List of all netCDF dimension and netCDF variable
#                   names
#
# global_attributes Set of global or non-standard CF attributes
#                   which have identical values across all input
#                   spaces.
#
# dimN              Integer
# auxN              Integer
# cmN               Integer
# dataN             Integer
# gmN               Integer
# bndN              Integer
# bnddimN           Integer
# strlenN           Integer
# ----------------- ------------------------------------------------------

# Initialize global variables
netcdf            = None
nc                = {}
dim_to_ncdim      = {}
ncdim_size        = {}
seen              = {} 
ncvar_names       = []
global_attributes = set(())
dimN              = 1
auxN              = 1
cmN               = 1
dataN             = 1
gmN               = 1
bndN              = 1
bnddimN           = 1
strlenN           = 1

def write(spaces, file, format='NETCDF3_CLASSIC', 
          data=False, verbose=False):
    '''

Write a space, or sequence of spaces, to a CF compliant netCDF file.

NetCDF dimension and variable names will be taken from variables'
private attributes `ncvar` and the grid attribute `nc_dimensions` if
present, otherwise they are inferred from standard names or set to
defaults. NetCDF names may be automatically given a numerical suffix
to avoid duplication.

Output netCDF file global attributes are those which occur in the set
of CF global attributes and non-standard data variable attributes and
which have equal values across all input spaces.

It does not matter if an input space's data has not yet been read into
memory. The write operation will automatically read the data array
from disk if required, but the default is for it not to be saved in
the input space if it had to be read during the write
operation. However, if the `data` argument equals True then any input
spaces which didn't contain their full data arrays in memory will be
changed in place so that they do. For example, for a space, s:

>>> s.type()
[<type 'netCDF4.Variable'>]
>>> write(s, 'filename')
>>> s.type()
[<type 'netCDF4.Variable'>]
>>> write(s, 'filename', data=True)
>>> s.type()
[<type 'numpy.ndarray'>]

Identical grid components (as determined by their numerically tolerant
`equals` methods) are only written to the file once, apart from when
they need to fulfil both dimension coordinate and auxiliary coordinate
roles for different data variables.


:argument spaces: The space or spaces to write to the file.

:type spaces: Space or sequence of spaces       

:argument file: The netCDF file to write spaces to.

:type file: str

:parameter data: *Optional*. If True then ensure that spaces which
                 have been written to disk retain their data in-place
                 as numpy arrays.

:type data: bool

:parameter format: *Optional*. The format of the output netCDF
                   file. Valid formats are those supported by the
                   netCDF4 module: "NETCDF3_CLASSIC",
                   "NETCDF3_64BIT", "NETCDF4_CLASSIC" and
                   "NETCDF4". The default is
                   "NETCDF3_CLASSIC". Certain write operations may
                   be considerably faster using "NETCDF4_CLASSIC".

:type format: str

:parameter verbose: *Optional*. If True then print information to
                    stdout.

:type verbose: bool

:returns: None


**Examples:**

>>> s
[<CF Space: air_pressure(30, 24)>,
 <CF Space: u_compnt_of_wind(19, 29, 24)>,
 <CF Space: v_compnt_of_wind(19, 29, 24)>,
 <CF Space: potential_temperature(19, 30, 24)>]
>>> write(s, 'file')
'''
    # See notes in this module for the definition of these global
    # variables
    global netcdf
    global nc
    global dim_to_ncdim                       
    global ncdim_size 
    global seen  
    global ncvar_names
    global dimN    
    global auxN    
    global cmN     
    global dataN   
    global gmN     
    global bndN    
    global bnddimN 
    global strlenN 
 
    def _grid_ncdimensions(f, key):
        '''
'''
        # Global dictionary
        global dim_to_ncdim
        
        if f.grid[key].ndim == 0:
            return ()
        else:
            return tuple(dim_to_ncdim[dim] for dim in f.grid.dimensions[key])
    #--- End: def
    
    def _grid_ncvar(s, key, default, counter):
        '''
'''
        ncvar = s.grid[key].getpriv('ncvar', s.grid[key].name(default=default))

        return _check_name(ncvar, counter)        
    #--- End: def
    
    def _create_dimension(ncdim, f, dim):
        '''

Create a netCDF dimension variable. Returns None.
'''        
        # Global dictionaries
        global dim_to_ncdim, ncdim_size
               
        ncdim_size[ncdim] = f.grid.dimension_sizes[dim]
        dim_to_ncdim[dim] = ncdim

        netcdf.createDimension(ncdim, ncdim_size[ncdim])
    #--- End: def

    def _create_dimension_coordinate(f, key):
        '''

Create a dimension coordinate variable, creating a new bounds
dimension if required, and create its bounds or climatology variable
if it has one. Returns seen[id(f.grid[key])]['ncvar']).

    '''       
        # Global dictionaries
        global nc, dim_to_ncdim, seen
        # Global counters
        global dimN

        create = False
        if not _seen(f.grid[key]):            
            create = True
        elif seen[id(f.grid[key])]['ncdims'] != ():
            if seen[id(f.grid[key])]['ncvar'] != seen[id(f.grid[key])]['ncdims'][0]:
                # Already seen this coordinate, but it had a different
                # netCDF dimension name
                create = True

        if create:
            ncvar, dimN = _grid_ncvar(f, key, 'dim', dimN)
            
            # Create a new dimension, if it is not a scalar coordinate
            if f.grid[key].ndim > 0:
                _create_dimension(ncvar, f, key)

            ncdimensions = _grid_ncdimensions(f, key)
            
            # Create a new dimension coordinate variable
            _createVariable(ncvar, ncdimensions, f.grid[key], data=data)
            
            # If this dimension coordinate has bounds then create the
            # bounds netCDF variable and add the bounds or climatology
            # attribute
            try:
                ncbounds = _create_bounds(f, key, ncdimensions)
            except AttributeError:
                pass
            else:
                try:
                    if f.grid[key].climatology:
                        nc[ncvar].setncattr('climatology', ncbounds)
                    else:
                        nc[ncvar].setncattr('bounds', ncbounds)
                except AttributeError:
                    nc[ncvar].setncattr('bounds', ncbounds)
            #--- End: try
        #--- End: if

        ncvar = seen[id(f.grid[key])]['ncvar']
        try:
            dim_to_ncdim[key] = ncvar
        except KeyError:
            pass

        return ncvar
    #--- End: def

    def _create_auxiliary_coordinate(s, key, coordinates):
        '''

Create an auxiliary coordinate variable and create its bounds or
climatology variable if it has one. Return the coordinates list
updated with the ncvar for the auxiliary coordinate.
    
    '''
        # Global dictionaries
        global nc, seen
        # Global counters
        global auxN

        ncdimensions = _grid_ncdimensions(s, key)
        if not _seen(s.grid[key], ncdimensions):
            
            ncvar, auxN = _grid_ncvar(s, key, 'aux', auxN)
            
            # Create a new auxiliary coordinate variable
            _createVariable(ncvar, ncdimensions, s.grid[key],
                            data=data)
            
            # If this auxiliary coordinate has bounds then create the
            # bounds netCDF variable and add the bounds or climatology
            # attribute
            try:
                ncbounds = _create_bounds(s, key, ncdimensions)
            except AttributeError:
                pass
            else:
                try:
                    if s.grid[key].climatology:
                        nc[ncvar].setncattr('climatology', ncbounds)
                    else:
                        nc[ncvar].setncattr('bounds', ncbounds)
                except AttributeError:
                    nc[ncvar].setncattr('bounds', ncbounds)
            #--- End: try
        #--- End: if

        coordinates.append(seen[id(s.grid[key])]['ncvar'])

        return coordinates
    #--- End: def

    def _create_bounds(s, key, coord_ncdimensions):
        '''

Create a coordinate's bounds netCDF variable, creating a new bounds
netCDF dimension if required. Return the bounds variable's netCDF
variable name.
'''
        # Global netCDF4.Dataset
        global netcdf
        # Global dictionaries
        global nc, seen, ncdim_size
        # Global counters
        global bndN, bnddimN
        
        try:
            bounds = s.grid[key].bounds
            bounds._data
        except AttributeError as error:
            raise AttributeError, error

        # Still here? Then this coordinate has a bounds attribute
        # which contains data.
        size = bounds.shape[-1]

        ncdim, bnddimN = _check_name('bounds%d' % size, bnddimN, size)

        # Check if this bounds variable has not been previously
        # created.
        ncdimensions = coord_ncdimensions +(ncdim,)        
        if _seen(bounds, ncdimensions):
            return seen[id(bounds)]['ncvar']    
        
        # Still here? Then this bounds variable has not been
        # previously created, so create it now.
        try:
            ncdim_size[ncdim]
        except KeyError:
            ncdim_size[ncdim] = size
            netcdf.createDimension(ncdim, ncdim_size[ncdim])

        ncvar = bounds.getpriv('ncvar', '%s_bounds' % s.grid[key].getpriv('ncvar'))
            
        ncvar, bndN = _check_name(ncvar, bndN)

        # Create the bounds netCDF variable
        _createVariable(ncvar, ncdimensions, bounds, data=data)

        return ncvar
    #--- End: def
    
    def _seen(variable, ncdims=None):
        '''

Return True if a cf variable identical to the input 'variable' (via
variable.equals) is already in the 'seen' dictionary (and has
therefore already been written to the output file), otherwise return
False. When True is returned, the 'seen' dictionary is updated for the
input 'variable'.

If 'ncdims' is set then a further condition for returning True is that
'ncdims' is equal to netCDF dimensions (names and order) of the
identical variable in the 'seen' dictionary.
'''
        # Global dictionary
        global seen

        for ident in seen:
            if ncdims is not None and ncdims != seen[ident]['ncdims']:
                continue

            if variable.equals(seen[ident]['variable']):
                seen[id(variable)] = {'variable': variable,
                                      'ncvar'   : seen[ident]['ncvar'],
                                      'ncdims'  : seen[ident]['ncdims']}
                return True
        #--- End: for

        return False
    #--- End: def

    def _datatype(variable):
        '''

Return the netCDF4.createVariable datatype corresponding to the
datatype of the array stored (possibly virtually) in the input
variable's 'data' attribute.

Numpy string data types will return 'S1' regardless of the numpy
string length. This means that the required conversion of
multi-character datatype numpy arrays into single-character datatype
numpy arrays (with an extra trailing dimension) is expected to be done
elsewhere (currently in def _createVariable).

If the input variable has no data attribute then 'S1' is returned.
'''
        try:
            variable.dtype
        except AttributeError:
            return 'S1'

        if variable.dtype == 'float32': return 'f4'
        if variable.dtype == 'float64': return 'f8'
        if variable.dtype == 'float'  : return 'f8'
        if variable.dtype == 'int32'  : return 'i4'
        if variable.dtype == 'int64'  : return 'i8'
        if variable.dtype == 'int8'   : return 'i1'
        if variable.dtype == 'int16'  : return 'i2'
        if variable.dtype is None     : return 'S1'        
        if variable.dtype.char == 'S' : return 'S1'            

        return None
    #--- End: def

    def _createVariable(ncvar, ncdimensions, input, omit=(),
                        data=False):
        '''

Create a netCDF variable from 'input' with name 'ncvar' and dimensions
'ncdimensions'. The netCDF variable's attributes are given by
input.properties(), less any given by the 'omit' argument. If a new
string-length netCDF dimension is required then it will also be
created. The 'seen' dictionary is updated for 'input'.

    '''
        # Global netCDF4.Dataset
        global netcdf
        # Global dictionaries
        global seen, nc, ncdim_size
        # Global counters
        global strlenN

        # Set the netCDF4.createVariable datatype
        datatype = _datatype(input)

        dimensions = ncdimensions

        data_out = None 
        if data:
            try:
                data_out = input.varray
            except AttributeError:
                pass            
        else:
            try:
                if not isinstance(input._data._list[0], numpy.ndarray):
#                    pointer = input._data
                    data_out = input.varray
#                    input._data = pointer
                else:
                    data_out = input.varray                    
            except AttributeError:
                pass
            

        # Convert a multi-character datatype numpy array into
        # single-character datatype numpy array with an extra trailing
        # dimension
        strlen = False
        if datatype == 'S1' and data_out is not None:
           strlen = int(re.findall('(\d+$)', str(input.dtype))[0])
           if strlen > 1:
               ncdim, strlenN = _check_name('strlen', strlenN, strlen)
               try:
                   ncdim_size[ncdim]
               except KeyError:
                   ncdim_size[ncdim] = strlen
               netcdf.createDimension(ncdim, ncdim_size[ncdim])
               dimensions = ncdimensions + (ncdim,)

               data_out = data_out.flatten()
               temp = numpy.empty((data_out.size, ncdim_size[ncdim]), '|S1')
               for i in xrange(data_out.size):
                   temp[i,:] = list(data_out[i].ljust(strlen, ' '))
               data_out = temp.reshape(input.shape + (ncdim_size[ncdim],))
               del temp
        #--- End: if

        # Create a new netCDF variable
        nc[ncvar] = netcdf.createVariable(ncvar, datatype, dimensions)

        # Add properties
        pubattrs = input.pubattrs
        [pubattrs.pop(attr) for attr in omit if attr in pubattrs]
	for attr, value in pubattrs.iteritems():
	    nc[ncvar].setncattr(attr, value)

        # Add data
        if data_out is not None:
            nc[ncvar][...] = data_out

        del data_out

        # Update the 'seen' dictionary
        seen[id(input)] = {'variable': input,
                           'ncvar'   : ncvar,
                           'ncdims'  : dimensions,
                           }
    #--- End: def

    def _create_grid_mapping(ncvar, input):
        '''
    
Create a netCDF grid_mapping variable from :obj:`input` with name
:obj:`ncvar`. The netCDF variable's attributes are given by key/value
pairs of input.iteritems(). The 'seen' dictionary is updated for
'input'.

    '''
        # Global netCDF4.Dataset
        global netcdf
        # Global dictionaries
        global seen, nc

        # Create a new netCDF variable
        nc[ncvar] = netcdf.createVariable(ncvar, 'S1', ())

        # Add attributes from key/value pairs
	for attr, value in input.iteritems():
	    nc[ncvar].setncattr(attr, value)

        # Update the 'seen' dictionary
        seen[id(input)] = {'variable': input, 
                           'ncvar'   : ncvar, 
                           'ncdims'  : (), # Transforms have no dimensions
                           }
    #--- End: def

    def _check_name(base, counter, dimsize=None):
        '''


'''
        # Global list
        global ncvar_names
        # Global dictionary
        global ncdim_size

        if dimsize is not None:
            if base in ncvar_names and dimsize == ncdim_size[base]:
                return base, counter

        if base in ncvar_names:
            ncvar = '%(base)s_%(counter)d' % locals()
            while ncvar in ncvar_names:
                counter += 1
                ncvar = '%(base)s_%(counter)d' % locals()
        else:
            ncvar = base

        ncvar_names.append(ncvar)            
        return ncvar, counter
    #--- End: def
    
    def _write_space(s, save=False):
        
        # Global netCDF4.Dataset
        global netcdf
        # Global sets
        global global_attributes
        # Global dictionaries
        global ncdim_size, dim_to_ncdim, nc, seen
        # Global counters
        global dataN, dimN, auxN, cmN, gmN

        if save:
            saved_dim_to_ncdim = copy.deepcopy(dim_to_ncdim)

        dim_to_ncdim = {}

        coordinates  = []

        for key in sorted(s.grid.dimension_sizes):            
            
            try:
                s.grid[key]
            except KeyError:
                # There is no dimension coordinate for this dimension,
                # so just create a netCDF dimension if required.
                try:
                    ncdim = s.grid.nc_dimensions[key]
                except (AttributeError, KeyError):
                    ncdim = 'dim'

                ncdim, dimN = _check_name(ncdim, dimN)         

                if (s.grid.dimension_sizes[key] > 1 or 
                    key in s.grid.dimensions['data']):
                    # Create the new dimension
                    _create_dimension(ncdim, s, key)

            else:              
                # A dimension coordinate exists for this dimension, so
                # create a netCDF dimension and a netCDF coordinate
                # netCDF variable. 
                ncdim = _create_dimension_coordinate(s, key)

                # Add scalar coordinates to the list of auxiliary
                # coordinates
                if s.grid[key].ndim == 0:
                    coordinates.append(ncdim)
            #--- End: try

        #--- End: for

                
        # ----------------------------------------------------------------------
        # Create auxiliary coordinate variables, except those which
        # might be completely specified elsewhere by a transformation.
        # ----------------------------------------------------------------------
        # Initialize the list of 'coordinates' attribute variable
        # values (each of the form 'name')
        for key in sorted(s.grid.get_keys('^aux')):
            coordinates = _create_auxiliary_coordinate(s, key, coordinates)
        #--- End: for

        # ----------------------------------------------------------------------
        # Create cell measures variables
        # ----------------------------------------------------------------------
        # Initialize the list of 'cell_measures' attribute values
        # (each of the form 'measure: name')
        cell_measures = []
        for key in sorted(s.grid.get_keys('^cm')):
            
            ncdimensions = _grid_ncdimensions(s, key)
            if not _seen(s.grid[key], ncdimensions):
                ncvar, cmN = _grid_ncvar(s, key, 'cm', cmN)        

                # Create a new cell_measures variable
                _createVariable(ncvar, ncdimensions, s.grid[key],
                                data=data)
                
            cell_measures.append('%s: %s' % (f.space[cm].standard_name,
                                             seen[id(f.space[key])]['ncvar']))
        #--- End: for

        # ----------------------------------------------------------------------
        # Create grid mapping and formula terms variables
        # ----------------------------------------------------------------------
        # Initialize the list of 'grid_mapping' attribute variable
        # values (each of the form 'name')
        grid_mapping  = []
        formula_terms = []
        try:
            transforms = s.grid.transform.values()
        except AttributeError:
            pass
        else:     

            for transform in transforms:
                
                try:
                    transform['grid_mapping_name']
                except KeyError:
                    # ------------------------------------------------
                    # This transform is a formula terms
                    # ------------------------------------------------

                    key = s.grid.coord(transform.name, exact=True, key=True)
                    if key is None:
                        # This formula_terms doesn't match up with an
                        # existing coordinate
                        continue
                    
                    # Still here? Then create the formula_terms
                    # attribute string
                    for term, var in transform.iteritems():
                        
                        if var is None:
                            # Do not create broken formula_terms
                            formula_terms = None
                            break
                    
                        if isinstance(var, Variable):                    
                            try:
                                seen[id(var)]
                            except KeyError:
                                # Value is a variable which hasn't
                                # been written yet (set save=True so
                                # we don't overwrite the global
                                # dim_to_ncdim dictionary).
                                _write_space(var, save=True)
            
                            ncvar = seen[id(var)]['ncvar']
                            formula_terms.append('%(term)s: %(ncvar)s' % locals())
                        else:
                            try:
                                ncvar = seen[id(s.grid[var])]['ncvar']
                                formula_terms.append('%(term)s: %(ncvar)s' % locals())
                            except :
                                # Do not create broken formula_terms
                                formula_terms = None
                                break
                       #--- End: if
                   #--- End: for
            
                    # Add the formula_terms attribute to the output
                    # variable
                    if formula_terms:
                        ncvar = seen[id(s.grid[key])]['ncvar']
                        nc[ncvar].setncattr('formula_terms', ' '.join(formula_terms))

                else:           
                    # ------------------------------------------------
                    # This transform is a grid mapping
                    # ------------------------------------------------
                    if _seen(transform):
                        ncvar = seen[id(transform)]['ncvar']
                    else:
                        # It has not yet been witten to the file, so
                        # create it
                        try:
                            ncvar = transform.ncvar
                        except AttributeError:
                            ncvar = grid_mapping_name
                        ncvar, gmN = _check_name(ncvar, gmN) 
                            
                        _create_grid_mapping(ncvar, transform)
                        
                        ncvar = seen[id(transform)]['ncvar']
                    #--- End: if
                    grid_mapping.append(ncvar)
                #--- End: try

            #--- End: for

        #--- End: try

        # ------------------------------------------------------------
        # Create the data variable
        # ------------------------------------------------------------
        ncvar = s.getpriv('ncvar', s.name(default='data'))
            
        ncvar, dataN = _check_name(ncvar, dataN)   
        
        ncdimensions = tuple(dim_to_ncdim[dim] 
                             for dim in s.grid.dimensions['data'])

        # Create a new data variable
        _createVariable(ncvar, ncdimensions, s,
                        omit=global_attributes | set(('ancillary_variables',
                                                      'flag_meanings',
                                                      'cell_methods',
                                                      )),
                        data=data)

        try:
            nc[ncvar].flag_meanings = ' '.join(nc[ncvar].flag_meanings)
        except AttributeError:
            pass
        
        if cell_measures:
            nc[ncvar].cell_measures = ' '.join(cell_measures)           
        if coordinates:
            nc[ncvar].coordinates   = ' '.join(coordinates)
        if grid_mapping:
            nc[ncvar].grid_mapping  = ' '.join(grid_mapping)

        # CELL METHODS
        try:
            cell_methods = copy.deepcopy(s.cell_methods)
        except AttributeError:
            pass
        else:
            for cm in cell_methods:
                for j in xrange(len(cm['dim'])):
                    dim = cm['dim'][j]
                    if dim not in (None, False) and not cm['no_coords'][j]:
                        cm['name'][j] = seen[id(s.grid[dim])]['ncvar']
            #--- End: for
            nc[ncvar].cell_methods = ' '.join(cell_methods.strings())

#            for i in xrange(len(cell_methods)):
#                for j in xrange(len(cell_methods[i]['dim'])):
#                    dim = cell_methods[i]['dim'][j]
#                    if dim not in (None, False) and not cell_methods[i]['no_coords'][j]:
#                        cell_methods[i]['name'][j] = seen[id(s.grid[dim])]['ncvar']
#            nc[ncvar].cell_methods = ' '.join(cell_methods.strings())
            
        # ------------------------------------------------------------
        # If we are saving the original dim_to_ncdim then copy back
        # the original
        # ------------------------------------------------------------
        if save:
            dim_to_ncdim = copy.deepcopy(saved_dim_to_ncdim)
    #--- End: def
            
    def _create_global_attributes(spaces):
        '''

Find the input spaces' global attributes and write them to the
netCDF4.Dataset. These are the global or non-standard CF attributes
which have identical values across all input spaces. Returns None.
'''
        # Global netCDF4.Dataset
        global netcdf
        # Global sets
        global global_attributes

        data_attributes = ('add_offset',
                           'ancillary_variables',
                           'cell_measures',
                           'cell_methods',
                           'coordinates',
                           '_FillValue',
                           'flag_masks',
                           'flag_meanings',
                           'flag_values',
                           'grid_mapping',
                           'long_name',
                           'missing_value',
                           'scale_factor',
                           'standard_error_multiplier',
                           'standard_name',
                           'units',
                           'valid_max',
                           'valid_min',
                           'valid_range',
                           )

        global_attributes = set(('comment',
                                 'Conventions',
                                 'history',
                                 'institution',
                                 'references',
                                 'source',
                                 'title',
                                 ))
        
        # Put all non-standard CF attributes (i.e. those not in the
        # data_attributes tuple) into the global_attributes list
        for s in spaces:
            for prop in set(s.pubattrs) - global_attributes:
                if prop not in data_attributes:
                    global_attributes.add(prop)
        #--- End: for

        # Remove attributes from the global_attributes set which have
        # different values in different data variables
        f0 = spaces[0]
        for attr in list(global_attributes):
            try:
                f0.getpub(attr)
            except AttributeError:
                global_attributes.remove(attr)
        #--- End: for

        if len(spaces) > 1:
            for attr in list(global_attributes):
                attr0 = f0.getpub(attr)
                for f in spaces[1:]:
                    try:                    
                        if not equals (f.getpub(attr), attr0):
                            global_attributes.remove(attr)
                            break
                    except AttributeError:
                        global_attributes.remove(attr)
                        break
            #--- End: for
        #--- End: if

        try:
            global_attributes.remove('Conventions')
        except KeyError:
            pass
        netcdf.setncattr('Conventions', 'CF-1.5')

        for attr in global_attributes:
            netcdf.setncattr(attr, f0.getpub(attr))

        # Put Conventions back into the global attributes set so that
        # we know to exclude it from data variables' attributes
        global_attributes.add('Conventions')
    #--- End: def

    # ------------------------------------------------------------
    # Reset global variables. See notes in this module for the
    # definition of these global variables
    # ------------------------------------------------------------
    nc                = {}
    dim_to_ncdim      = {}
    ncdim_size        = {}
    seen              = {} 
    ncvar_names       = []
    global_attributes = set(())
    dimN              = 1
    auxN              = 1
    cmN               = 1
    dataN             = 1
    gmN               = 1
    bndN              = 1
    bnddimN           = 1
    strlenN           = 1

    # ---------------------------------------------------------------
    # Open the output netCDF file. (NETCDF3_CLASSIC can be *really*
    # slow)
    # ---------------------------------------------------------------
    netcdf = netCDF4.Dataset(file, 'w', format=format)

    # ---------------------------------------------------------------
    # Set the fill mode for a Dataset open for writing to off. This
    # will prevent the data from being pre-filled with fill values,
    # which may result in some performance improvements.
    # ---------------------------------------------------------------
    netcdf.set_fill_off()

    # ---------------------------------------------------------------
    # Write global attributes to the file first. This is important as
    # doing it later slows things down enormously. This function
    # creates the global global_attributes set, which is used in def
    # _write_space.
    # ---------------------------------------------------------------
    _create_global_attributes(spaces)

    # ---------------------------------------------------------------
    # ---------------------------------------------------------------
    for space in spaces:
        _write_space(space)
        if verbose:
            print repr(space)
        
    # ---------------------------------------------------------------
    # Close the netCDF Dataset, writing all buffered data to disk
    # ---------------------------------------------------------------
    netcdf.close()

    if verbose:
        n = len(spaces)
        print 'Wrote',repr(n),'space'+('s' if n>1 else '')+' to',file
#--- End: def
